#donotrun

/*
Simple 3D setup with anti-alias and DOF.

Implement this function in fragments that include this file:
vec3 color(vec3 cameraPos, vec3 direction);

*/

#buffer RGBA32F

#buffershader "BufferShader-1.0.1.frag"
#info Simple 3D Setup
#camera 3D

#vertex

#group Camera
// Field-of-view
uniform float FOV; slider[0,0.4,2.0] NotLockable
uniform vec3 Eye; slider[(-50,-50,-50),(0,0,-10),(50,50,50)] NotLockable
uniform vec3 Target; slider[(-50,-50,-50),(0,0,0),(50,50,50)] NotLockable
uniform vec3 Up; slider[(0,0,0),(0,1,0),(0,0,0)] NotLockable

varying vec3 from;
uniform vec2 pixelSize;
varying vec2 coord;
varying vec2 viewCoord;
varying vec2 viewCoord2;
varying vec3 dir;
varying vec3 Dir;
varying vec3 UpOrtho;
varying vec3 Right;
uniform int subframe;
uniform int backbufferCounter;
varying vec2 PixelScale;

#ifdef providesInit
void init(); // forward declare
#else
void init() {}
#endif

void main(void)
{
	gl_Position =  gl_Vertex;
	coord = (gl_ProjectionMatrix*gl_Vertex).xy;
	coord.x*= pixelSize.y/pixelSize.x;
	// we will only use gl_ProjectionMatrix to scale and translate, so the following should be OK.
	PixelScale =vec2(pixelSize.x*gl_ProjectionMatrix[0][0], pixelSize.y*gl_ProjectionMatrix[1][1]);
	viewCoord = gl_Vertex.xy;
	viewCoord2= (gl_ProjectionMatrix*gl_Vertex).xy;

	from = Eye;
	Dir = normalize(Target-Eye);
	UpOrtho = normalize( Up-dot(Dir,Up)*Dir );
	Right = normalize( cross(Dir,UpOrtho));
	coord*=FOV;
	init();
}
#endvertex

#group Camera
uniform bool EquiRectangular; checkbox[false]

#group Raytracer

#define PI  3.14159265358979323846264
#define TWO_PI  6.28318530717958647692529

// Camera position and target.
varying vec3 from;
varying vec3 dir;
varying vec3 dirDx;
varying vec3 dirDy;
varying vec2 coord;
varying float zoom;

//uniform int subframe;
uniform int backbufferCounter;
uniform sampler2D backbuffer;
varying vec2 viewCoord;
varying vec2 viewCoord2;
varying vec3 Dir;
varying vec3 UpOrtho;
varying vec3 Right;
uniform float FocalPlane; slider[0,1,5]
uniform float Aperture; slider[0,0.00,0.2]

uniform int ApertureNbrSides; slider[2,7,10] 
//Rotation of the diaphragm
uniform float ApertureRot; slider[0,0,360]
//For star shaped diphragms. Very limited
uniform bool ApStarShaped; checkbox[false] 
uniform float BokehCircle; slider[0,1.7,5]

//#define subframe backbufferCounter 
#ifdef OLD_FRAG
int subframe = backbufferCounter; 
#else
uniform int subframe;
#endif

/*
vec2 rand2(vec2 co){
	// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
	return
	vec2(fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453),
		fract(cos(dot(co.xy ,vec2(4.898,7.23))) * 23421.631));
}
*/
#ifdef WANGHASH
//Wang hash by Syntopia
//http://www.fractalforums.com/index.php?topic=22721.msg88910#msg88910
#extension GL_ARB_shader_bit_encoding : enable
#extension GL_EXT_gpu_shader4 : enable
#extension GL_ARB_gpu_shader5 : enable
uint wang_hash(uint seed)
{
	seed = (seed ^ 61u) ^ (seed >> 16u);
	seed *= 9u;
	seed = seed ^ (seed >> 4u);
	seed *= 0x27d4eb2du;
	seed = seed ^ (seed >> 15u);
	return seed ;
}
// Wrapper for getting from float to ints. This certainly looses precision. I imagine we could do better here. //float wang_hash_fp(vec2 pos)
float rand(vec2 pos){
	uint ix = floatBitsToUint(pos.x);
	uint iy = floatBitsToUint(pos.y);
    // I use two hash calls to untangle pos.x and pos.y
	return float(wang_hash(wang_hash(ix)+iy)) / 4294967296.0;
}
float rand(float v)
{
        uint ix = floatBitsToUint(v);
        return float(wang_hash(ix)) / 4294967296.0;
}

//for iq-clouds
float rand(vec3 pos){
	uint ix = floatBitsToUint(pos.x);
	uint iy = floatBitsToUint(pos.y);
	uint iz = floatBitsToUint(pos.z);
	return float(wang_hash(wang_hash(wang_hash(ix)+iy)+iz)) / 4294967296.0;
}

vec2 rand2(vec2 pos){
	uint ix = floatBitsToUint(pos.x);
	uint iy = floatBitsToUint(pos.y);
    // I use two hash calls to untangle pos.x and pos.y
	return vec2(float(wang_hash(wang_hash(ix)+iy)) / 4294967296.0, float(wang_hash(wang_hash(iy)+ix)) / 4294967296.0);
}
#else
float rand(float v) {
// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
	return fract(sin(dot(v ,78.233)) * 43758.5453);
}
float rand(vec2 pos) {
// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
    return fract(sin(dot(pos.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

vec2 rand2(vec2 co){
	// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
	return
	vec2(fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453),
		fract(cos(dot(co.xy ,vec2(4.898,7.23))) * 23421.631));
}
#endif
//Modified from Syntopia's. http://blog.hvidtfeldts.net/index.php/2015/01/path-tracing-3d-fractals/
vec2 seed = viewCoord*(float(subframe)+1.0);
vec2 rand2n() {//there are too much rand versions out there. I'll need to clean things up.
    seed+=vec2(-1.0,1.0);
    //seed+=vec2(wang_hash_fp(seed.x),wang_hash_fp(seed.y));
	return rand2(seed);
}

vec2 NGonDisc(vec2 co) {
	vec2 r = rand2(co);
	float d=1./tan(PI/float(ApertureNbrSides));
	float lh=1./sqrt(1.+d*d);
	vec2 p=vec2(r.x+r.y, r.x-r.y);
	if(!ApStarShaped) p.x=1.-abs(p.x-1.);
	p.x*=d;
	p*=lh;
	float a=dot(rand2(11.*co+vec2(1.,1.)),vec2(1.,1.));
	a=floor(a*float(ApertureNbrSides))*2.*PI/float(ApertureNbrSides) + ApertureRot*PI/360.;
	vec2 cs= vec2(cos(a), sin(a));
	p= vec2(cs.x*p.x-cs.y*p.y, cs.x*p.y+cs.y*p.x);
	/*a=ApertureRot*PI/360.;
	cs= vec2(cos(a), sin(a));
	p= vec2(cs.x*p.x-cs.y*p.y, cs.x*p.y+cs.y*p.x);*/
	return p;
}

vec2 uniformDisc(vec2 co) {
	vec2 r = rand2(co);
	//return sqrt(r.y)*vec2(cos(r.x*6.28),sin(r.x*6.28));
	float r2 = pow((1.-pow(r.y,BokehCircle)), 1./BokehCircle);
	return r2*vec2(cos(r.x*TWO_PI),sin(r.x*TWO_PI));
}


#ifdef providesInit
void init(); // forward declare
#else
void init() {}
#endif

//out vec4 gl_FragColor;

#group Post
uniform float Gamma; slider[0.0,2.0,5.0];
// 1: Linear, 2: Expontial, 3: Filmic, 4: Reinhart; 5: Syntopia
uniform int ToneMapping; slider[1,4,5];
uniform float Exposure; slider[0.0,1.0,3.0];
uniform float Brightness; slider[0.0,1.0,5.0];
uniform float Contrast; slider[0.0,1.0,5.0];
uniform float Saturation; slider[0.0,1.0,5.0];
uniform float GaussianWeight; slider[0.0,1.0,10.0];

uniform float AntiAliasScale; slider[0,2,10];

varying vec2 PixelScale;
uniform float FOV;

// implement this;
vec3 color(vec3 cameraPos, vec3 direction);

// Given a camera pointing in 'dir' with an orthogonal 'up' and 'right' vector
// and a point, coord, in screen coordinates from (-1,-1) to (1,1),
// a ray tracer direction is returned
vec3 equiRectangularDirection(vec2 coord, vec3 dir, vec3 up, vec3 right)  {
	vec2 r = vec2(coord.x,(1.0-coord.y)*0.5)*PI;
	return cos(r.x)*sin(r.y)*dir+
		 sin(r.x)*sin(r.y)*right+
		cos(r.y)*up;
}

void main() {
	init();
	// A Thin Lens camera model with Depth-Of-Field
	// We want to sample a circular diaphragm
	// Notice: this is not sampled with uniform density
	vec2 r = vec2(0.);
	if(ApertureNbrSides == 2)
		// We want to sample a circular diaphragm
		r = Aperture*uniformDisc(viewCoord*(float(subframe)+1.0));
	else
		// We want to sample a polygonal diaphragm
		r = Aperture*NGonDisc(viewCoord*(float(subframe)+1.0));	
	vec2 disc = uniformDisc(coord*float(subframe+1)); // subsample jitter
	vec2 jitteredCoord;
	// Direction from Lens positon to point on FocalPlane
	vec3 lensOffset = r.x*Right + r.y*UpOrtho;
        vec3 rayDir;
	if (EquiRectangular) {
	  jitteredCoord = AntiAliasScale*PixelScale*disc;
	}
	else {
	  jitteredCoord = coord + AntiAliasScale*PixelScale*FOV*disc;
	}
	//vec3 rayDir = normalize (Dir+ jitteredCoord.x*Right+jitteredCoord.y*UpOrtho)*FocalPlane-(lensOffset);
	rayDir = (Dir+ jitteredCoord.x*Right+jitteredCoord.y*UpOrtho)*FocalPlane-(lensOffset);
	rayDir = normalize(rayDir);
	if (EquiRectangular) {
	  rayDir = equiRectangularDirection(viewCoord2, rayDir, UpOrtho, Right);
	}
	vec3 c =  color(from+lensOffset,rayDir);
	// Accumulate
	vec4 prev = texture2D(backbuffer,(viewCoord+vec2(1.0))*0.5);
	float w =1.0-length(disc);
	if (GaussianWeight>0.) {
		w = exp(-dot(disc,disc)/GaussianWeight)-exp(-1.0/GaussianWeight);
	}

	gl_FragColor = (prev+ vec4(c*w, w));

}
