#version 420
#info Theli-at's Pseudo Kleinian (Scale 1 JuliaBox + Something (here a Menger Sponge)).
#include "MathUtils.frag"
//#include "iqPath_Light.frag"
#include "newIQPath.frag"

#group PseudoKleinian

// Made by Knighty, see this thread:
// http://www.fractalforums.com/3d-fractal-generation/fragmentarium-an-ide-for-exploring-3d-fractals-and-other-systems-on-the-gpu/msg32270/#msg32270

#define USE_INF_NORM

// Maximum iterations
uniform int MI; slider[0,5,20]

// Size
uniform float Size; slider[0,1,2]

// Cubic fold Size
uniform vec3 CSize; slider[(0,0,0),(1,1,1),(2,2,2)]

// Julia constant
uniform vec3 C; slider[(-2,-2,-2),(0,0,0),(2,2,2)]

// Basic shape DE Offset
uniform float DEoffset; slider[0,0,0.01]

// Basic shape Translation
uniform vec3 Offset; slider[(-2,-2,-2),(0,0,0),(2,2,2)]

//Basic shape
// Menger Number of iterations.
uniform int MnIterations;  slider[0,2,20]
uniform int ColorIterations;  slider[0,2,20]

// Menger Scale parameter. A perfect Menger is 3.0
uniform float MnScale; slider[0.00,3.0,4.00]

// Menger Scaling center
uniform vec3 MnOffset; slider[(0,0,0),(1,1,1),(2,2,2)]

float Menger(vec3 z)
{
	float r;
	
	int n = 0;
	// Fold
	z = abs(z);
	if (z.x<z.y){ z.xy = z.yx;}
	if (z.x<z.z){ z.xz = z.zx;}
	if (z.y<z.z){ z.yz = z.zy;}
	if (z.z<1./3.){ z.z -=2.*( z.z-1./3.);}
	
while (n < MnIterations && dot(z,z)<100.0) {
		
		z=MnScale* (z-MnOffset)+MnOffset;
		
		// Fold
		z = abs(z);
		if (z.x<z.y){ z.xy = z.yx;}
		if (z.x< z.z){ z.xz = z.zx;}
		if (z.y<z.z){ z.yz = z.zy;}
		if (z.z<1./3.*MnOffset.z){ z.z -=2.*( z.z-1./3.*MnOffset.z);}
		
		r = dot(z-MnOffset, z-MnOffset);
		//orbitTrap = min(orbitTrap, abs(vec4(z,r)));
		
		n++;
	}
	
	//return (length(z)-sqrt(3.) ) * pow(Scale, float(-n));
	return float(z.x-MnOffset) * pow(MnScale, float(-n));
}

float Thing2(vec3 p){
//Just scale=1 Julia box
#ifdef USE_INF_NORM   
	vec3 p1=abs(p);
	float r2=max(p1.x,max(p1.y,p1.z));
#else
	float r2=dot(p,p);
#endif
	float DEfactor=1.;
   
	for(int i=0;i<MI && r2<60.;i++){
		p=2.*clamp(p, -CSize, CSize)-p;
      
		r2=dot(p,p);
		//orbitTrap = min(orbitTrap, abs(vec4(p,r2)));
		float k=max(Size/r2,1.);
		p*=k;DEfactor*=k;
      
		p+=C;
#ifdef USE_INF_NORM   
		p1=abs(p);
		r2=max(p1.x,max(p1.y,p1.z));
#else
		 r2=dot(p,p);
#endif
		//if (i < ColorIterations) orbitTrap = min(orbitTrap, abs(vec4(p,r2)));
	}
	//Call basic shape and scale its DE
	return abs(0.5*Menger(p-Offset)/DEfactor-DEoffset);
}

float DE(vec3 p){
	return  Thing2(p);//RoundBox(p, CSize, Offset);
}






#preset Default
FOV = 0.50406
Eye = -2.35627,-2.98332,-9.72622
Target = 3.4429,-11.0423,-8.53323
Up = 0.0449125,-0.114574,-0.992296
EquiRectangular = false
FocalPlane = 1.1598
Aperture = 0
ApertureNbrSides = 7
ApertureRot = 0
ApStarShaped = false
BokehCircle = 1.7
Gamma = 1.4
ToneMapping = 2
Exposure = 1.10205
Brightness = 1
Contrast = 1.1386
Saturation = 1
GaussianWeight = 1
AntiAliasScale = 2
RayBounces = 2
RaySteps = 383
FudgeFactor = 1
HitDistance = 0.00076
SoftShadows = 1
MaxDepth = 42.2567
Mat_Color = 0.882353,0.843137,0.623529
Mat_Refl = 0
Mat_ReflExp = 0.73171
UseTexture = false
DeTex = 
DeTexScale = 0.5
DeTexExposure = 1
OrbitStrength = 0
X = 0.5,0.6,0.6,-0.20388
Y = 1,0.6,0,0.5534
Z = 0.8,0.78,1,0.4369
R = 0.4,0.7,1,0.86274
CycleColors = true
Cycles = 6.1755
SpotLight = 1,0.611765,0.294118,3
SpotLightDir = 0.70128,-0.12344
SpotSize = 0.01
SpotExp = 0.5
BgColor = 0.298039,0.498039,0.698039,7
Ambience = 0.541176,0.541176,0.541176,4.4262
Fog = 0.288,0,0,0.39063
Fog_Exp = 0.66667
GRIter = 11
GRPower = 0.48485
UseFloor = true
FloorColor = 0.196078,0.388235,0.639216
FloorReflection = 1
FloorReflectionExp = 1
FloorHeight = 9.02
FloorRot = -92.0916,0,180
MI = 5
Size = 1
CSize = 1,1,1
C = 0,1,0
DEoffset = 0
Offset = 0.07408,0.62964,0
MnIterations = 8
ColorIterations = 5
MnScale = 1.73076
MnOffset = 0.95918,1.71428,1
#endpreset



