// Created by inigo quilez - iq/2016 https://www.shadertoy.com/view/Xtt3Wn
// License Creative Commons Attribution-NonCommercial-ShareAlike 3.0
// More info here: http://iquilezles.org/www/articles/simplepathtracing/simplepathtracing.htm

//adapt for Fragmentaium by CJR (^_^)
//all rands in 3D-1.frag
#info define WANGHASH for best quality
//need speculars, soft shadows, id's, material mode...

#include "3D-1.frag" 

#group Raytracer
uniform int RayBounces;slider[1,5,128]
uniform int RaySteps;slider[64,128,4000]
uniform float FudgeFactor;slider[0.1,1.0,1.0]
uniform float HitDistance;slider[0.00001,0.00018,0.001]
uniform float SoftShadows;slider[1.0,1.0,250.0]
uniform float MaxDepth;slider[1,15,200]

#group Material
uniform vec3 Mat_Color; color[.65,.65,.65]
uniform float Mat_Refl;slider[0.0,0.0,1.0]
uniform float Mat_ReflExp;slider[0.00001,.75,1.0]
uniform bool UseTexture; checkbox[false]
uniform sampler2D DeTex; file[tex.jpg]
uniform float DeTexScale;slider[0.0,0.5,10.0]
uniform float DeTexExposure;slider[0.0,1.0,2.0]

#group Coloring
vec4 orbitTrap = vec4(10000.0);
uniform float OrbitStrength; slider[0,0,1]
uniform vec4 X; color[-1,0.7,1,0.5,0.6,0.6];
uniform vec4 Y; color[-1,0.4,1,1.0,0.6,0.0];
uniform vec4 Z; color[-1,0.5,1,0.8,0.78,1.0];
uniform vec4 R; color[-1,0.12,1,0.4,0.7,1.0];
uniform bool CycleColors; checkbox[false]
uniform float Cycles; slider[0.1,1.1,32.3]

#group Environment
uniform vec4 SpotLight; color[0.0,1.,10.0,1.0,.8,.5];
uniform vec2 SpotLightDir;  slider[(-2,-2),(0.1,0.1),(2,2)]
uniform float SpotSize;slider[0.0,0.01,10.0]
uniform float SpotExp;slider[0.0001,0.5,20.0]
uniform vec4 BgColor; color[0.0,1.,10.0,.3,.5,.7];
uniform vec4 Ambience; color[0.0,1.,10.0,.3,.5,.7];
//uniform float Ambience;slider[0.0,1.0,5.0]
uniform vec4 Fog; color[0.0,0.0,1.5,.3,.5,.7];
uniform float Fog_Exp;slider[0.00001,.25,2.0]

#group GodRays
uniform int GRIter;slider[0,3,20]
uniform float GRPower; slider[0.0,.3,1.0]

#group Floor
uniform bool UseFloor; checkbox[False]
uniform vec3 FloorColor;color[0.45,0.64,0.53]
uniform float FloorReflection;slider[0.0,0.0,1.0]
uniform float FloorReflectionExp;slider[0.0,1.0,1.0]
uniform float FloorHeight;slider[0.0,1.0,2.0]
uniform vec3 FloorRot; slider[(-180,-180,-180),(0,0,0),(180,180,180)]

mat4 frot;

struct material {vec3 color; float refl, reflExp;};
material Mater0=material(Mat_Color,Mat_Refl,Mat_ReflExp);
material Mater1=material(FloorColor,FloorReflection,FloorReflectionExp);
material Mater2=material(vec3(0.6,0.4,0.1),1.0,1.0);

vec3 cosineDirection(in vec3 nor)
{//return a random direction on the hemisphere
	vec2 r = rand2n()*TWO_PI;
	vec3 dr=vec3(sin(r.x)*vec2(sin(r.y),cos(r.y)),cos(r.x));
	return (dot(dr,nor)<0.0)?-dr:dr;
}

vec3 uniformVector( in float seed)
{
    float a = PI*rand( 78.233 + seed);
    float b = TWO_PI*rand( 10.873 + seed);
    return vec3( sin(b)*sin(a), cos(b)*sin(a), cos(a) );
}

vec3 cycle(vec3 c, float s) {
	return vec3(0.5)+0.5*vec3(cos(s*Cycles+c.x),cos(s*Cycles+c.y),cos(s*Cycles+c.z));
}
vec3 getColor() {
	orbitTrap.w = sqrt(orbitTrap.w);
	
	vec3 orbitColor;
	if (CycleColors) {
		orbitColor = cycle(X.xyz,orbitTrap.x)*X.w*orbitTrap.x +
		cycle(Y.xyz,orbitTrap.y)*Y.w*orbitTrap.y +
		cycle(Z.xyz,orbitTrap.z)*Z.w*orbitTrap.z +
		cycle(R.xyz,orbitTrap.w)*R.w*orbitTrap.w;
	} else {
		orbitColor = X.xyz*X.w*orbitTrap.x +
		Y.xyz*Y.w*orbitTrap.y +
		Z.xyz*Z.w*orbitTrap.z +
		R.xyz*R.w*orbitTrap.w;
	}
	
	vec3 color = mix(Mat_Color, 3.0*orbitColor,  OrbitStrength);
	return color;
}

vec4 texcube( sampler2D sam, in vec3 p, in vec3 n )
{
    vec3 m = n*n;
	vec4 x = texture2D( sam, p.yz );
	vec4 y = texture2D( sam, p.zx );
	vec4 z = texture2D( sam, p.xy );
	return (x*m.x + y*m.y + z*m.z) / (m.x+m.y+m.z);
}

vec2 min2(vec2 d1, vec2 d2){return (d1.x<d2.x)?d1:d2;}//sorts vectors based on .x

float DE(vec3 pos);
float DESphere(in vec3 z, float radius){return length(z)-radius;}

vec2 map(in vec3 z)
{//return distance estimate and object id
	vec2 de=vec2(DE(z),0.0);//start your object ids at zero
		frot = rotationMatrix(vec3(1.0,0.0,0.0), FloorRot.x)*rotationMatrix(vec3(0.0,1.0,0.0), FloorRot.y)*rotationMatrix(vec3(0.0,0.0,1.0), FloorRot.z);
		vec4 p = vec4(z,0.0)*frot;
	vec2 flr=(UseFloor==true)?vec2((p.y-FloorHeight),1.0):vec2(1,2.0);//add as many ids as you likee
	vec2 bal=vec2(DESphere(z+vec3(0.0,0.15,-0.5),0.3),2.0);
	return min2(min2(de,flr),bal);//add as many objects as you like this way
}

material getMaterial( in vec3 z0, in vec3 nor, in float item )
{//get material properties 
	Mater0.color=getColor();
	if(item==0.0) return Mater0;
	if(item==1.0) return Mater1; 
	if(item==2.0) return Mater2; 
}


//vec3 calcNormal( in vec3 pos )
vec3 calcNormal( in vec3 pos, in float item )
{
    vec3 eps = vec3(0.000001,0.0,0.0);
    return normalize( vec3(
      map( pos+eps.xyy ).x - map( pos-eps.xyy ).x,
      map( pos+eps.yxy ).x - map( pos-eps.yxy ).x,
      map( pos+eps.yyx ).x - map( pos-eps.yyx ).x ) );
}


vec2 intersect(in vec3 ro, in vec3 rd )
{
	float t=HitDistance*10.0,d=1000.0;
	for(int i=0;i<RaySteps && abs(d)>HitDistance;i++){ 
		t+=d=map(ro+t*rd).x*FudgeFactor;
		if(t>=MaxDepth) return vec2(t,-1.0);//-1.0 is id for "hit nothing"
	}
	vec2 h=map(ro+t*rd);//move close to the hit point without fudging
	h.x=t+h.x;
	return h;//returns distance, object id
}


float shadow( in vec3 ro, in vec3 rd)
{
	float res = 0.0;
	float tmax = MaxDepth;
	float t = 0.000001;
	for(int i=0;i<RaySteps;i++){	
		float h = map(ro + rd*t).x;
        if( h<HitDistance || t>tmax) break;
        t += h;
    }
	if( t>tmax ) res = 1.0;
    return res;
}



vec3 color(vec3 ro, vec3 rd)
{
    //const float epsilon = 0.0001;
	float epsilon = HitDistance;
    vec3 colorMask = vec3(1.0);
    vec3 accumulatedColor = vec3(0.0);
    vec3 oro = ro;
    vec3 ord = rd;
    float fdis = 0.0;
	vec2 t = vec2(0.0);
	float sa = rand(viewCoord*(float(subframe)+1.0));
	vec3 sunDir = vec3(sin(SpotLightDir.x*PI)*cos(SpotLightDir.y*PI/2.0), sin(SpotLightDir.y*PI/2.0)*sin(SpotLightDir.x*PI), cos(SpotLightDir.x*PI));
	sunDir = normalize(sunDir);
	vec3 sunCol = SpotLight.xyz*SpotLight.w; 
	vec3 skyCol=(Fog.w>0)?Fog.xyz*Fog.w+BgColor.xyz*BgColor.w:BgColor.xyz*BgColor.w;
	if(SpotLight.w!=0)skyCol+=pow(max(0.0,dot(rd,sunDir))*1.001  ,SpotExp*100)*SpotSize*SpotLight.xyz;
	
    for( int bounce = 0; bounce<RayBounces; bounce++ ) // bounces of GI
    {
        rd = normalize(rd);
        // trace
		t = intersect( ro, rd );
        if( t.y < 0.0 ){
		
            if( bounce==0 ){ 
				accumulatedColor=skyCol;
				break;
			}
            break;
        }
        if( bounce==0 ) fdis = t.x;
        vec3 pos = ro + rd * t.x;
        vec3 nor = calcNormal( pos, t.y );
		material m=getMaterial( pos, nor, t.y );
		vec3 surfaceColor = clamp(m.color,0.0,1.0);
        if(UseTexture)surfaceColor+=clamp(texcube(DeTex, pos*DeTexScale, nor ).xyz,0.0,1.0)*DeTexExposure;
       
        // add direct lighitng
        vec3 iColor = vec3(0.0);
		colorMask *= surfaceColor;
		// light 1        
        float sunDif =  max(0.0, dot(sunDir, nor));
        float sunSha = 1.0; if( sunDif > 0.00001 ) sunSha = shadow( pos + nor*epsilon, sunDir);
        iColor += sunCol * sunDif * sunSha;
        // todo - add back direct specular
        // light 2
		vec3 skyPoint = cosineDirection(nor);
        float skySha = shadow( pos + nor*epsilon, skyPoint);
		iColor += Ambience.xyz*skySha*Ambience.w;
        accumulatedColor += colorMask * iColor;
		
		/*
		float acc = 0.0;
		for( int i=0; i<GRIter; i++ ){
		float t = fdis*rand(sa+13.731*float(i)+float(subframe)*7.773);
		vec3 pos = ro + rd*t;
		acc += GRPower*shadow( pos, sunDir );//*exp(-0.25*t);	
		}
		accumulatedColor += vec3(0.1)*pow(acc,1.0)*sunCol;//*0.4;
		*/
		
        // calculate new ray	
		float isDif = 1.-m.refl;
        if( rand(sa + 1.123 + 7.7*float(bounce)) < isDif ){
			rd = cosineDirection(nor);//rd = cosineDirection(76.2 + 73.1*float(bounce) + sa + 17.7*float(subframe), nor);
			//rd = normalize(rd);
        } else {
			rd=normalize(reflect(rd, nor))+uniformVector(sa + 111.123 + 65.2*float(bounce))*(1.-m.reflExp);
			
        }
        ro = pos;
   }
   float ff = exp(-0.1*fdis);
   accumulatedColor *= ff; 
 
   // volumetrics
   float acc = 0.0;
   vec3 pos = oro + ord;
   for( int i=0; i<GRIter; i++ ) {
	   float t = fdis*rand(sa+13.731*float(i)+float(subframe)*7.773);
       vec3 pos = oro + ord*t;
       acc += GRPower*shadow( pos, sunDir );//*exp(-0.25*t);
	}
    accumulatedColor += vec3(0.1)*pow(acc,2.0)*sunCol*0.4;
	float fg=1.0*Fog_Exp-exp(-pow(Fog.w,4.0)*fdis*fdis)*Fog_Exp; //simple fog
	skyCol+=Fog.xyz;
	accumulatedColor = mix(accumulatedColor, skyCol, fg);
return accumulatedColor;
}
